/*
Code to control LivingColors-lights by using a CC2500.
*/


#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "init.h"
#include "cc2500.h"
#include "rgbtohsv.h"
#include "convtab.h"

#define CS	PB2
#define SI	PB3
#define SO	PB4
#define CLK	PB5

volatile unsigned char rgbdata[9];
volatile char update_rgb=3;


//Routine used by the UART reception routines to actually set the RGB values
//of the lights.
inline void setrgb(unsigned char* dat) {
    int x;
    static int t;
    char u=0;

    //See if anything has changed
    for (x=0; x<9; x++) {
	if (rgbdata[x]!=dat[x]) u=1;
    }

    //Change anyway every 15 non-changing data packets
    if (u==0) t++;
    if (t>15) u=1;

    if (u) {
	for (x=0; x<9; x++) rgbdata[x]=dat[x];
	update_rgb=3;
	t=0;
    }

}

//The adresses of the LivingColors-lights to control. Gets filled in
//with data from the EEPROM later.
unsigned char addrs[][10]={
    {0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
    {0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
    {0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00}};

//Write addresses of lamps to the EEPROM
void writecfg() {
    int x,y;
    char p=0;
    eeprom_write_byte(p++,0x55);
    for (y=0; y<3; y++) {
	for (x=0; x<11; x++) {
	    eeprom_busy_wait();
	    eeprom_write_byte(p++,addrs[y][x]);
	}
    }
}

//Read adressed from EEPROM
void readcfg() {
    int x,y;
    char p=0;
    //Refuse to read uninited eeprom
    if (eeprom_read_byte(p++)!=0x55) return;
    for (y=0; y<3; y++) {
	for (x=0; x<11; x++) {
	    addrs[y][x]=eeprom_read_byte(p++);
	}
    }

}

//Button is pressed: put the CC2500 in listen mode for a while and snoop codes
//sent by the remote control.
//Learn the code in 'code', if not known yet.
void learncode(unsigned char* code) {
    //Check if code is known
    int x,y;
    char found,codeissame;
    found=0;
    for (y=0; y<3; y++) {
	codeissame=1;
	if (addrs[y][0]==0) codeissame=0;
	for (x=0; x<10; x++) {
	    if (code[x]!=addrs[y][x+1]) codeissame=0;
	}
	if (codeissame) found=1;
    }
    //Found = 1 means we already got that code. Exit if that's the case.
    if (found) return;

    //Find new slot to store code
    y=0;
    while (addrs[y][0]==1 && y<3) y++;
    if (y==3) return; //full
    
    //store
    addrs[y][0]=1;
    for (x=0; x<10; x++) {
	addrs[y][x+1]=code[x];
    }
}

//The LivingColors idea of hue differs a bit from the hue everyone else uses;
//while the r->y->g->c->b->m->r-idea persists, it's not completely
//linear. The non-linearity is solved by looking the hue to send up in a
//lookuptable defined in convtab.h.
unsigned char adjust_hue_for_livcol(unsigned char realhue) {
    return pgm_read_byte(hue_convtab+realhue);
}

void update_lamps(unsigned char code) {
    update_rgb=3;
    update_a_lamp(code);
}


//This routine is called periodically and sends the rgb-values of to exactly one
//lamp. The function-code used to do this is configurable:
//Codes: 3-set color, 7-turn lamp off, 5-turn lamp on
void update_a_lamp(unsigned char code) {
    static char lamp;
    static char seq;
    int j;
    unsigned char h,s,v;


    if (update_rgb>0) {
	//led to indicate we're sending
    	PORTC|=(1<<LED1);
	if (addrs[lamp][0]!=0) { //slot used?
	    seq++;
	    //convert rgb to hsv
	    rgb_to_hsv(rgbdata[lamp],rgbdata[lamp+3],rgbdata[lamp+6],&h,&s,&v);
	    //wait for a prev packet to be completely sent
	    disablecs();
	    while((ccsendbyte(0xF5)&0x1F)>1) {
	        disablecs();
	    }
	    //send byte
	    ccsendbyte(0x36);
	    disablecs();
	    ccsendbyte(0x7F);
	    ccsendbyte(0x0E);
	    for (j=0; j<9; j++) {
	        ccsendbyte(addrs[lamp][j+1]);
	    }
	    ccsendbyte(code);
	    ccsendbyte(seq++);
	    ccsendbyte(adjust_hue_for_livcol(h));
	    ccsendbyte(s);
	    ccsendbyte(v);
	    disablecs();
	    ccsendbyte(0x35);
	    disablecs();
	}
	lamp++; //next time, handle next lamp
	if (lamp>2) lamp=0; //only 3 lamps

	//sending done, indicate on led
	PORTC&=~(1<<LED1);
	update_rgb--; //one less lamp to update
    }
}


inline void disablecs(void) {
    PORTB|=(1<<CS);
//    _delay_us(150);
    _delay_us(60);
}

//Send a byte to the CC2500 serially
unsigned char ccsendbyte(unsigned char n) {
    char t;
    char r;
    if (PORTB|(1<<CS)) {
	PORTB&=~(1<<CS);
//	_delay_us(150);
	_delay_us(60);
    }
    r=0;
    for (t=0; t<8; t++) {
	if (n&0x80) PORTB|=(1<<SO); else PORTB&=~(1<<SO);
	PORTB&=~(1<<CLK);
//	_delay_us(2);
	PORTB|=(1<<CLK);
	r<<=1;
	if (PINB&(1<<SI)) r|=1;
//	_delay_us(2);
	n<<=1;
    }
    PORTB&=~(1<<CLK);
    return r;
}

//Inits the CC2500 (and, incidentally, the rest of the board hardware)
void initcc2500(void) {
    short i;
    DDRB=(1<<CS)|(1<<SO)|(1<<CLK);
    DDRC|=(1<<LED1)|(1<<LED2)|(1<<LED3)|(1<<LED4);
    disablecs();
    _delay_ms(10);
    //Send cc2500 configuration data
    for (i=0; i<sizeof(init_cc2500); i++) {
	if ((i&3)==0) {
	    disablecs();
	    _delay_us(150);
	    PORTB|=(1<<CS);
	    _delay_us(150);
	}

	ccsendbyte(pgm_read_byte(init_cc2500+i));
	_delay_ms(1);
    }
    disablecs();
    ccsendbyte(0x36);
    disablecs();
    ccsendbyte(0x36);
    disablecs();

    //goto chan 1
    ccsendbyte(0x0A);
    ccsendbyte(1);
    disablecs();
    
    readcfg();
}


//Called when button is pressed. Should copy strings and addresses from 
//an existing remote controller.
void GoLearnCodes(void) {
    unsigned char code[17];
    char recved=0;
    unsigned char t;
    int x,l;
    disablecs();
    //Clear codes
    for (x=0; x<3; x++) addrs[x][0]=0;

    //Go receive
    while(recved<15) {
	disablecs();
	ccsendbyte(0x36);
	ccsendbyte(0x34);
	ccsendbyte(0x34);
	//wait for packet
	do {
	    disablecs();
	    ccsendbyte(0xF8);
	} while ((ccsendbyte(0)&1)==0);
	//fetch packet
	disablecs();
	ccsendbyte(0xff);
	l=ccsendbyte(0);
	for (x=0; x<l; x++) {
	    t=ccsendbyte(0);
	    if (x<17) code[x]=t;
	}
	disablecs();
	//Check for correct len and turn-off packet
	if (l==0xe && code[9]==7) {
	    learncode(code);
	    recved++;
	}
    }
    disablecs();
    ccsendbyte(0x36);
    ccsendbyte(0x36);
    disablecs();
    ccsendbyte(0x36);
    ccsendbyte(0x36);
    disablecs();
    
    //enable lights
/*
    l=0x80;
    for (l=0; l<255; l++) {
	//set rgb, force update
	for (x=0; x<9; x++) rgbdata[x]=l;
	update_rgb=1;
	//turn found lamps on
	update_lamps(5);
//	_delay_ms(2);
    }
*/

    //write addresses to eeprom
    writecfg();
}
